"""
Segmentation using Fully-Convolutional Network

References
----------
[1] https://github.com/pytorch/vision/tree/master/torchvision/models/segmentation

"""

# from collections import OrderedDict
# from copy import deepcopy
# from numbers import Number, Real
# from typing import Any, List, NoReturn, Optional, Sequence, Tuple, Union

# import numpy as np
# import pandas as pd
# import torch
# import torch.nn.functional as F
# from torch import Tensor, nn
