# -*- coding: utf-8 -*-
"""
docstring, to write
"""

from .cpsc2018 import CPSC2018
from .cpsc2019 import CPSC2019
from .cpsc2020 import CPSC2020
from .cpsc2021 import CPSC2021

__all__ = [
    "CPSC2018",
    "CPSC2019",
    "CPSC2020",
    "CPSC2021",
]
